@@ -15,6 +15,12 @@
  * You should have received a copy of the GNU General Public License
  * along with this program.  If not, see <http://www.gnu.org/licenses/>.
  *----------------------------------------------------------------------------
+ * Version 1.6
+ *   - Zero transparent pixels to guard against invisible color information
+ *     (Slight slowdown, but more robustness)
+ *   - Add SSE2 runtime detection
+ *   - Correctly set DropFrame attribute
+ *
  * Version 1.5
  *   - Validate framerates
  *   - Make non-drop timecodes the default for 23.976, 29.97 and 59.94
@@ -48,6 +54,7 @@
 #include <stdint.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <time.h>
 #include <png.h>
 
 /* AVIS input code taken from muxers.c from the x264 project (GPLv2 or later).
@@ -244,18 +251,30 @@ void write_png(int file_id, uint8_t *image, int w, int h)
 int is_identical_sse2 (stream_info_t *s_info, char *img, char *img_old)
 {
 	char *max = img + s_info->i_width * s_info->i_height * 4;
+	char zero[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
+	char a_mask[16] = {0,0,0,0xff,0,0,0,0xff,0,0,0,0xff,0,0,0,0xff};
+	char full[16] = {0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff};
 	int volatile different = 0;
-	int *diff_ptr = &different;
+	int volatile *diff_ptr = &different;
 
+	asm __volatile__ ("movupd (%%eax), %%xmm3\n" : : "a" (zero));
+	asm __volatile__ ("movupd (%%ecx), %%xmm4\n" : : "c" (a_mask));
+	asm __volatile__ ("movupd (%%edx), %%xmm7\n" : : "d" (full));
 	while (img < max && !different)
 	{
 		asm __volatile__
 		(
 			"movapd (%%eax), %%xmm0\n"
+			"movapd %%xmm0, %%xmm5\n"
+			"pand %%xmm4, %%xmm5\n"
 			"movapd (%%esi), %%xmm1\n"
+			"pcmpeqd %%xmm3, %%xmm5\n"
+			"pxor %%xmm7, %%xmm5\n"
+			"pand %%xmm5, %%xmm0\n"
+			"movapd %%xmm0, (%%eax)\n"
 			"psadbw %%xmm1, %%xmm0\n"
 			"movd %%xmm0, (%%edx)\n"
-			: "=d" (diff_ptr)
+			: "=a" (img), "=d" (diff_ptr)
 			: "a" (img), "S" (img_old), "d" (diff_ptr)
 		);
 		img += 16;
@@ -272,15 +291,18 @@ int is_empty_sse2 (stream_info_t *s_info, char *img)
 {
 	char *max = img + s_info->i_width * s_info->i_height * 4;
 	char zero[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
+	char a_mask[16] = {0,0,0,0xff,0,0,0,0xff,0,0,0,0xff,0,0,0,0xff};
 	int volatile not_empty = 0;
-	int *ne_ptr = &not_empty;
+	int volatile *ne_ptr = &not_empty;
 
-	asm __volatile__ ("movupd (%%ebx), %%xmm1\n" : : "b" (zero));
+	asm __volatile__ ("movupd (%%eax), %%xmm1\n" : : "a" (zero));
+	asm __volatile__ ("movupd (%%ecx), %%xmm2\n" : : "c" (a_mask));
 	while (img < max && !not_empty)
 	{
 		asm __volatile__
 		(
 			"movapd (%%esi), %%xmm0\n"
+			"pand %%xmm2, %%xmm0\n"
 			"psadbw %%xmm1, %%xmm0\n"
 			"movd %%xmm0, (%%edx)\n"
 			: "=d" (ne_ptr)
@@ -295,6 +317,34 @@ int is_empty_sse2 (stream_info_t *s_info, char *img)
 	return 1;
 }
 
+void zero_transparent_sse2 (stream_info_t *s_info, char *img)
+{
+	char *max = img + s_info->i_width * s_info->i_height * 4;
+	char zero[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
+	char a_mask[16] = {0,0,0,0xff,0,0,0,0xff,0,0,0,0xff,0,0,0,0xff};
+	char full[16] = {0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff};
+
+	asm __volatile__ ("movupd (%%eax), %%xmm1\n" : : "a" (zero));
+	asm __volatile__ ("movupd (%%ecx), %%xmm2\n" : : "c" (a_mask));
+	asm __volatile__ ("movupd (%%edx), %%xmm3\n" : : "d" (full));
+	while (img < max)
+	{
+		asm __volatile__
+		(
+			"movapd (%%eax), %%xmm0\n"
+			"movapd %%xmm0, %%xmm5\n"
+			"pand %%xmm2, %%xmm5\n"
+			"pcmpeqd %%xmm1, %%xmm5\n"
+			"pxor %%xmm3, %%xmm5\n"
+			"pand %%xmm5, %%xmm0\n"
+			"movapd %%xmm0, (%%eax)\n"
+			: "=a" (img)
+			: "a" (img)
+		);
+		img += 16;
+	}
+}
+
 void swap_rb_sse2 (stream_info_t *s_info, char *img, char *out)
 {
 	char *max = img + s_info->i_width * s_info->i_height * 4;
@@ -340,24 +390,44 @@ int is_identical_c (stream_info_t *s_info, char *img, char *img_old)
 	uint32_t *im_old = (uint32_t *)img_old;
 
 	while (im < max)
+	{
+		if (!((char *)im)[3])
+			*im = 0;
 		if (*(im++) ^ *(im_old++))
 			return 0;
+	}
 
 	return 1;
 }
 
 int is_empty_c (stream_info_t *s_info, char *img)
 {
-	uint32_t *max = (uint32_t *)(img + s_info->i_width * s_info->i_height * 4);
-	uint32_t *im = (uint32_t *)img;
+	char *max = img + s_info->i_width * s_info->i_height * 4;
+	char *im = img;
 
 	while (im < max)
-		if (*(im++))
+	{
+		if (im[3])
 			return 0;
+		im += 4;
+	}
 
 	return 1;
 }
 
+void zero_transparent_c (stream_info_t *s_info, char *img)
+{
+	char *max = img + s_info->i_width * s_info->i_height * 4;
+	char *im = img;
+
+	while (im < max)
+	{
+		if (!im[3])
+			*(uint32_t *)img = 0;
+		im += 4;
+	}
+}
+
 void swap_rb_c (stream_info_t *s_info, char *img, char *out)
 {
 	char *max = img + s_info->i_width * s_info->i_height * 4;
@@ -373,31 +443,65 @@ void swap_rb_c (stream_info_t *s_info, char *img, char *out)
 	}
 }
 
+int detect_sse2 ()
+{
+	static int detection = -1;
+	unsigned int func = 0x00000001;
+	unsigned int eax, ebx, ecx, edx;
+
+	if (detection != -1)
+		return detection;
+
+	asm volatile
+	(
+		"cpuid\n"
+		: "=a" (eax), "=b" (ebx), "=c" (ecx), "=d" (edx)
+		: "a" (func)
+	);
+
+	/* SSE2:  edx & 0x04000000
+	 * SSSE3: ecx & 0x00000200
+	 */
+	detection = (edx & 0x04000000) ? 1 : 0;
+
+	if (detection)
+		fprintf(stderr, "CPU: Using SSE2 optimized functions.\n");
+	else
+		fprintf(stderr, "CPU: Using pure C functions.\n");
+
+	return detection;
+}
+
 int is_identical (stream_info_t *s_info, char *img, char *img_old)
 {
-#ifdef HAVE_SSE2
-	return is_identical_sse2(s_info, img, img_old);
-#else
-	return is_identical_c(s_info, img, img_old);
-#endif
+	if (detect_sse2())
+		return is_identical_sse2(s_info, img, img_old);
+	else
+		return is_identical_c(s_info, img, img_old);
 }
 
 int is_empty (stream_info_t *s_info, char *img)
 {
-#ifdef HAVE_SSE2
-	return is_empty_sse2(s_info, img);
-#else
-	return is_empty_c(s_info, img);
-#endif
+	if (detect_sse2())
+		return is_empty_sse2(s_info, img);
+	else
+		return is_empty_c(s_info, img);
+}
+
+void zero_transparent (stream_info_t *s_info, char *img)
+{
+	if (detect_sse2())
+		zero_transparent_sse2(s_info, img);
+	else
+		zero_transparent_c(s_info, img);
 }
 
 void swap_rb (stream_info_t *s_info, char *img, char *out)
 {
-#ifdef HAVE_SSE2
-	swap_rb_sse2(s_info, img, out);
-#else
-	swap_rb_c(s_info, img, out);
-#endif
+	if (detect_sse2())
+		swap_rb_sse2(s_info, img, out);
+	else
+		swap_rb_c(s_info, img, out);
 }
 
 /* SMPTE time code */
@@ -478,24 +582,28 @@ struct framerate_entry_s
 	char *name;
 	char *out_name;
 	double rate;
+	int drop;
 };
 
 /* Most of the time seems to be spent in AviSynth (about 4/5). */
 int main (int argc, char *argv[])
 {
-	struct framerate_entry_s framerates[] = {{"23.976", "23.976", 24}, {"23.976d", "23.976", 24000/1001.0}, {"24", "24", 24}, {"25", "25", 25}, {"29.97", "29.97", 30}, {"29.97d", "29.97", 30000/1001.0}, {"50", "50", 50}, {"59.94", "59.94", 60}, {"59.94d", "59.94", 60000/1001.0}, {NULL, NULL, 0}};
+	struct framerate_entry_s framerates[] = {{"23.976", "23.976", 24, 0}, {"23.976d", "23.976", 24000/1001.0, 1}, {"24", "24", 24, 0}, {"25", "25", 25, 0}, {"29.97", "29.97", 30, 0}, {"29.97d", "29.97", 30000/1001.0, 1}, {"50", "50", 50, 0}, {"59.94", "59.94", 60, 0}, {"59.94d", "59.94", 60000/1001.0, 1}, {NULL, NULL, 0, 0}};
 	char *avs_filename, *track_name, *language, *video_format, *frame_rate, *out_filename;
 	char *in_img, *old_img, *tmp, *out_buf;
 	char *intc_buf, *outtc_buf;
+	char *drop_frame;
 	double fps;
 	int frames;
 	int first_frame = -1, start_frame, end_frame;
 	int num_of_events = 0;
 	int i;
 	int have_line = 0;
+	int must_zero;
 	int checked_empty;
 	int auto_cut = 0;
 	int progress_step = 1000;
+	int bench_start = time(NULL);
 	avis_input_t *avis_hnd;
 	stream_info_t *s_info = malloc(sizeof(stream_info_t));
 	event_list_head_t *events = malloc(sizeof(event_list_head_t));
@@ -516,15 +624,37 @@ int main (int argc, char *argv[])
 
 	/* TODO: Sanity check video_format and frame_rate. */
 
+	/* Get frame rate */
+	frame_rate = NULL;
+	i = 0;
+	while (framerates[i].name != NULL)
+	{
+		if (!strcasecmp(framerates[i].name, argv[5]))
+		{
+			frame_rate = framerates[i].out_name;
+			fps = framerates[i].rate;
+			drop_frame = framerates[i].drop ? "true" : "false";
+		}
+		i++;
+	}
+	if (frame_rate == NULL)
+	{
+		fprintf(stderr, "Error: Invalid framerate (%s).\n", argv[5]);
+		return 1;
+	}
+
+	/* Detect CPU features */
+	detect_sse2();
+
 	/* Get video info and allocate buffer */
 	if (open_file_avis(avs_filename, &avis_hnd, s_info))
 	{
 		print_usage();
 		abort();
 	}
-	in_img  = calloc(s_info->i_width * s_info->i_height * 4 + 16 * 4, sizeof(char)); /* allocate + 16 for alignment, and + n * 16 for over read/write */
-	old_img = calloc(s_info->i_width * s_info->i_height * 4 + 16 * 4, sizeof(char)); /* see above */
-	out_buf = calloc(s_info->i_width * s_info->i_height * 4 + 16 * 4, sizeof(char));
+	in_img  = calloc(s_info->i_width * s_info->i_height * 4 + 16 * 2, sizeof(char)); /* allocate + 16 for alignment, and + n * 16 for over read/write */
+	old_img = calloc(s_info->i_width * s_info->i_height * 4 + 16 * 2, sizeof(char)); /* see above */
+	out_buf = calloc(s_info->i_width * s_info->i_height * 4 + 16 * 2, sizeof(char));
 
 	/* align buffers */
 	in_img  = in_img + (16 - ((int)in_img % 16));
@@ -551,31 +681,12 @@ int main (int argc, char *argv[])
 			progress_step = 1;
 	}
 
-	/* Get frame rate */
-	frame_rate = NULL;
-	i = 0;
-	while (framerates[i].name != NULL)
-	{
-		if (!strcasecmp(framerates[i].name, argv[5]))
-		{
-			frame_rate = framerates[i].out_name;
-			fps = framerates[i].rate;
-		}
-		i++;
-	}
-	if (frame_rate == NULL)
-	{
-		fprintf(stderr, "Error: Invalid framerate (%s).\n", argv[5]);
-		return 1;
-	}
-
 	/* Initialize events */
 	events->start = NULL;
 	events->end = NULL;
 
-#ifdef HAVE_SSE2
-	asm ("emms\n");
-#endif
+	if (detect_sse2())
+		asm ("emms\n");
 
 	/* Process frames */
 	for (i = 0; i < frames; i++)
@@ -605,6 +716,9 @@ int main (int argc, char *argv[])
 		/* Check for duplicate, unless first frame */
 		if (i && have_line && is_identical(s_info, in_img, old_img))
 			continue;
+		/* Mark frames that were not used as new image in comparison to have transparent pixels zeroed */
+		else if (!(i && have_line))
+			must_zero = 1;
 
 		/* Not a dup, write end-of-line, if we had a line before */
 		if (have_line)
@@ -618,6 +732,11 @@ int main (int argc, char *argv[])
 		if (!checked_empty && is_empty(s_info, in_img))
 			continue;
 
+		/* Zero transparent pixels, if needed */
+		if (must_zero)
+			zero_transparent(s_info, in_img);
+		must_zero = 0;
+
 		/* Not an empty frame, start line */
 		have_line = 1;
 		start_frame = i;
@@ -650,9 +769,8 @@ int main (int argc, char *argv[])
 		abort();
 	}
 
-#ifdef HAVE_SSE2
-	asm ("emms\n");
-#endif
+	if (detect_sse2())
+		asm ("emms\n");
 
 	/* Initialize timecode buffers */
 	intc_buf = calloc(12, 1);
@@ -674,8 +792,8 @@ int main (int argc, char *argv[])
 		"<Description>\n"
 		"<Name Title=\"%s\" Content=\"\"/>\n"
 		"<Language Code=\"%s\"/>\n"
-		"<Format VideoFormat=\"%s\" FrameRate=\"%s\" DropFrame=\"false\"/>\n"
-		"<Events LastEventOutTC=\"%s\" FirstEventInTC=\"%s\"\n", track_name, language, video_format, frame_rate, outtc_buf, intc_buf);
+		"<Format VideoFormat=\"%s\" FrameRate=\"%s\" DropFrame=\"%s\"/>\n"
+		"<Events LastEventOutTC=\"%s\" FirstEventInTC=\"%s\"\n", track_name, language, video_format, frame_rate, drop_frame, outtc_buf, intc_buf);
 
 	mk_timecode(0, fps, intc_buf);
 	mk_timecode(frames, fps, outtc_buf);
@@ -710,5 +828,8 @@ int main (int argc, char *argv[])
 	fclose(fh);
 	close_file_avis(avis_hnd);
 
+	/* Give runtime */
+	/*fprintf(stderr, "Time elapsed: %d\n", time(NULL) - bench_start);*/
+
 	return 0;
 }
